﻿// PerspectiveControl.cs (c) 2009 by Charles Petzold
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace TextAtAngles
{
    public class PerspectiveControl : ContentControl
    {
        Matrix3DProjection projection = new Matrix3DProjection();

        public static readonly DependencyProperty BaseRectangleProperty =
            DependencyProperty.Register("BaseRectangle",
                typeof(Rect),
                typeof(PerspectiveControl),
                new PropertyMetadata(OnDependencyPropertyChanged));

        public static readonly DependencyProperty TopLeftProperty =
            DependencyProperty.Register("TopLeft",
                typeof(Point),
                typeof(PerspectiveControl),
                new PropertyMetadata(OnDependencyPropertyChanged));

        public static readonly DependencyProperty TopRightProperty =
            DependencyProperty.Register("TopRight",
                typeof(Point),
                typeof(PerspectiveControl),
                new PropertyMetadata(OnDependencyPropertyChanged));

        public static readonly DependencyProperty BottomLeftProperty =
            DependencyProperty.Register("BottomLeft",
                typeof(Point),
                typeof(PerspectiveControl),
                new PropertyMetadata(OnDependencyPropertyChanged));

        public static readonly DependencyProperty BottomRightProperty =
            DependencyProperty.Register("BottomRight",
                typeof(Point),
                typeof(PerspectiveControl),
                new PropertyMetadata(OnDependencyPropertyChanged));

        public PerspectiveControl()
        {
            Projection = projection;
        }

        public Rect BaseRectangle
        {
            set { SetValue(BaseRectangleProperty, value); }
            get { return (Rect)GetValue(BaseRectangleProperty); }
        }

        public Point TopLeft
        {
            set { SetValue(TopLeftProperty, value); }
            get { return (Point)GetValue(TopLeftProperty); }
        }

        public Point TopRight
        {
            set { SetValue(TopRightProperty, value); }
            get { return (Point)GetValue(TopRightProperty); }
        }

        public Point BottomLeft
        {
            set { SetValue(BottomLeftProperty, value); }
            get { return (Point)GetValue(BottomLeftProperty); }
        }

        public Point BottomRight
        {
            set { SetValue(BottomRightProperty, value); }
            get { return (Point)GetValue(BottomRightProperty); }
        }

        static void OnDependencyPropertyChanged(DependencyObject sender,
                                                DependencyPropertyChangedEventArgs args)
        {
            (sender as PerspectiveControl).OnDependencyPropertyChanged(args);
        }

        void OnDependencyPropertyChanged(DependencyPropertyChangedEventArgs args)
        {
            projection.ProjectionMatrix = CalculateNewTransform(BaseRectangle, TopLeft, TopRight, BottomLeft, BottomRight);
        }

        Matrix3D CalculateNewTransform(Rect rect, Point ptTopLeft, Point ptTopRight, 
                                                  Point ptBottomLeft, Point ptBottomRight)
        {
            // Scale and translate transform
            Matrix3D S = new Matrix3D()
            {
                M11 = 1 / rect.Width,
                M22 = 1 / rect.Height,
                OffsetX = -rect.Left / rect.Width,
                OffsetY = -rect.Top / rect.Height
            };

            // See http://www.charlespetzold.com/blog/2007/08/250638.html for math that follows

            // Affine transform
            Matrix3D A = new Matrix3D()
            {
                OffsetX = ptTopLeft.X,
                OffsetY = ptTopLeft.Y,
                M11 = (ptTopRight.X - ptTopLeft.X),
                M12 = (ptTopRight.Y - ptTopLeft.Y),
                M21 = (ptBottomLeft.X - ptTopLeft.X),
                M22 = (ptBottomLeft.Y - ptTopLeft.Y)
            };

            // Non-Affine transform
            Matrix3D B = new Matrix3D();
            double den = A.M11 * A.M22 - A.M12 * A.M21;
            double a = (A.M22 * ptBottomRight.X - A.M21 * ptBottomRight.Y + A.M21 * A.OffsetY - A.M22 * A.OffsetX) / den;
            double b = (A.M11 * ptBottomRight.Y - A.M12 * ptBottomRight.X + A.M12 * A.OffsetX - A.M11 * A.OffsetY) / den;

            B.M11 = a / (a + b - 1);
            B.M22 = b / (a + b - 1);
            B.M14 = B.M11 - 1;
            B.M24 = B.M22 - 1;

            // Product of three transforms
            return S * B * A;
        }

    }
}
